#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Provides cli helpers that can be used for usage with the local proxy in scripts.
"""
import logging
import time
import socket
import errno
from contextlib import contextmanager

import click


@contextmanager
def socketcontext(*args, **kw):
    """Socket context to ensure the socket is closed automatically when leaving the context"""
    sock = socket.socket(*args, **kw)
    try:
        yield sock
    finally:
        sock.close()


def wait_for_port(port: int, timeout: float = 30.0, silent: bool = False) -> None:
    """Wait for a port to be opened

    Args:
        port (int): Port number
        timeout (float): Timeout in seconds. Defaults to 30.0

    Raises:
        TimeoutError: Timeout error when the timeout period is exceeded
    """
    start_at = time.monotonic()
    expires_at = start_at + timeout
    while True:
        if is_port_open(port):
            duration = time.monotonic() - start_at
            if not silent:
                logging.info("Port is open. waited=%.3fs", duration)
            break

        if time.monotonic() >= expires_at:
            raise TimeoutError(
                f"Timed out waiting for port to be opened. port={port}, timeout={timeout}"
            )
        time.sleep(0.25)


def is_port_open(port: int, host: str = "127.0.0.1") -> bool:
    """Check if a port is open or not

    Args:
        port (int): Port number
        host (str): Host name

    Returns:
        bool: True if the port is open
    """
    with socketcontext() as sock:
        try:
            sock.bind((host, port))
        except OSError as ex:
            # port already in use error
            if ex.errno == errno.EADDRINUSE:
                return True

    return False


def get_unused_port() -> int:
    """Get a port which is currently unused"""
    with socketcontext() as sock:
        sock.bind(("127.0.0.1", 0))
        return sock.getsockname()[1]


CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    """Helper commands"""


@cli.command(name="wait")
@click.argument("port", type=click.IntRange(0, 65535))
@click.argument("timeout", type=click.FloatRange(0.0, 300.0))
@click.option("--silent", is_flag=True, help="Don't print out any error messages")
@click.pass_context
def cli_wait(ctx: click.Context, port, timeout, silent):
    """
    Wait for a port to be open

        \b
        PORT Port number to wait for
        TIMEOUT Timeout in seconds

    \b
    Example 1: Wait for port 2222 be open, but give up after 30 seconds

        \b
        python3 -m c8ylp.helper wait 2222 30

    """
    try:
        wait_for_port(port, timeout, silent)
    except TimeoutError:
        if not silent:
            click.secho(f"Port is not open after {timeout}s", fg="red")
        ctx.exit(1)


@cli.command(name="port")
def cli_port():
    """Get an unused port number

    Please note it does not guarantee that the port will not be taken by another application

    \b
    Example 1: Get a random unused port

        \b
        python3 -m c8ylp.helper port

    """
    click.echo(get_unused_port())


if __name__ == "__main__":
    cli()



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Unix Stream Socket server"""


import os
import logging
import threading
import socketserver

from c8ylp.tcp_socket.tcp_server import TCPProxyServer, TCPHandler


class CustomUnixStreamServer(socketserver.UnixStreamServer):
    """Custom TCP Server used to listen for local connections and proxy them to
    a websocket
    """

    def __init__(
        self,
        web_socket_client,
        server_address,
        RequestHandlerClass,
        bind_and_activate=True,
        buffer_size: int = 1024,
        tcp_timeout: float = 30,
    ) -> None:
        self.allow_reuse_address = True
        self.timeout = tcp_timeout
        self.buffer_size = buffer_size
        self.web_socket_client = web_socket_client
        super().__init__(
            server_address, RequestHandlerClass, bind_and_activate=bind_and_activate
        )


class UnixStreamProxyServer(TCPProxyServer):
    """
    Unix domain socket server
    """

    # pylint: disable=super-init-not-called
    def __init__(
        self,
        path,
        web_socket_client,
        tcp_buffer_size,
        tcp_timeout,
    ):
        self.web_socket_client = web_socket_client
        self._running = threading.Event()
        self.logger = logging.getLogger(__name__)
        self.path = path

        # Expose func to web socket client
        self.web_socket_client.proxy = self

        self.server = CustomUnixStreamServer(
            self.web_socket_client,
            path,
            TCPHandler,
            buffer_size=tcp_buffer_size,
            tcp_timeout=tcp_timeout,
        )

    def shutdown(self):
        """Shutdown tcp server"""
        super().shutdown()
        os.remove(self.path)



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Exit codes"""

import dataclasses
import logging
import os
import pathlib
import signal
import threading
import time
import sys
import socket
from enum import IntEnum
from logging.handlers import RotatingFileHandler
from typing import Any, Dict, NoReturn, Optional

import click

from ..timer import CommandTimer
from ..banner import BANNER1
from ..env import save_env
from ..rest_client.c8yclient import CumulocityClient, CumulocityMissingTFAToken
from ..tcp_socket import TCPProxyServer, StdioProxyServer
from ..websocket_client import WebsocketClient

if hasattr(socket, "AF_UNIX"):
    from ..tcp_socket import UnixStreamProxyServer


class ExitCodes(IntEnum):
    """Exit codes"""

    OK = 0
    NO_SESSION = 2
    NOT_AUTHORIZED = 3

    DEVICE_MISSING_REMOTE_ACCESS_FRAGMENT = 5
    DEVICE_NO_PASSTHROUGH_CONFIG = 6
    DEVICE_NO_MATCHING_PASSTHROUGH_CONFIG = 7
    MISSING_ROLE_REMOTE_ACCESS_ADMIN = 8

    UNKNOWN = 9

    SSH_NOT_FOUND = 10
    TIMEOUT_WAIT_FOR_PORT = 11
    COMMAND_NOT_FOUND = 12
    PLUGIN_EXECUTION_ERROR = 20
    PLUGIN_INVALID_FORMAT = 21
    PLUGIN_NOT_FOUND = 22
    TERMINATE = 100


@dataclasses.dataclass
class ProxyContext:
    """Local proxy context"""

    host = ""
    device = ""
    external_type = ""
    config = ""
    tenant = ""
    user = ""
    token = ""
    password = ""
    tfa_code = ""
    port = 0
    ping_interval = 0
    kill = False
    tcp_size = 0
    tcp_timeout = 0
    verbose = False
    ignore_ssl_validate = False
    ssh_user = ""
    additional_args = None
    disable_prompts = False
    env_file = None
    store_token = False
    wait_port_timeout = 60.0
    socket_path = ""
    stdio = False

    def __init__(self, ctx: click.Context, src_dict: Dict[str, Any] = None) -> None:
        self._ctx = ctx
        if src_dict is not None:
            self.fromdict(src_dict)

            configure_logger(CliLogger.log_path(), self.verbose)

    @property
    def _root_context(self) -> click.Context:
        return self._ctx.find_root().ensure_object(dict)

    @property
    def used_port(self) -> int:
        """Get the port used by the local proxy

        Returns:
            int: Port number
        """
        return self._root_context.get("used_port", self.port)

    @used_port.setter
    def used_port(self, value: int):
        """Store the port used by the local proxy for later reference

        Args:
            value (int): Port number
        """
        self._root_context["used_port"] = value

    def exit_server_not_ready(self) -> NoReturn:
        """Exit with a server not ready error

        Returns:
            NoReturn: The function does not return
        """
        self.show_error(
            "Timed out waiting for local port to open: "
            f"port={self.used_port}, timeout={self.wait_port_timeout}s"
        )
        self._ctx.exit(ExitCodes.TIMEOUT_WAIT_FOR_PORT)

    def fromdict(self, src_dict: Dict[str, Any]) -> "ProxyContext":
        """Load proxy settings from a dictionary

        Args:
            src_dict (Dict[str, Any]): [description]

        Returns:
            ProxyContext: Proxy options after the values have been set
                via the dictionary
        """
        logging.info("Loading from dictionary")
        assert isinstance(src_dict, dict)
        for key, value in src_dict.items():
            logging.info("reading key: %s=%s", key, value)
            if hasattr(self, key):
                setattr(self, key, value)
        return self

    def start_background(self, ctx: click.Context = None) -> "ProxyContext":
        """Start the local proxy in the background

        Returns:
            ProxyContext: Reference to the proxy context so it can be chained
                with other commands or used after the initialization of the class.
        """
        cur_ctx = ctx or self._ctx
        connection_data = pre_start_checks(cur_ctx, self)
        ready_signal = threading.Event()
        run_proxy_in_background(
            cur_ctx, self, connection_data=connection_data, ready_signal=ready_signal
        )
        if not ready_signal.wait(self.wait_port_timeout):
            self.exit_server_not_ready()
        return self

    def start(self, ctx: click.Context = None) -> None:
        """Start the local proxy in the background

        Returns:
            ProxyContext: Reference to the proxy context so it can be chained
                with other commands or used after the initialization of the class.
        """
        cur_ctx = ctx or self._ctx
        connection_data = pre_start_checks(cur_ctx, self)
        start_proxy(cur_ctx, self, connection_data=connection_data)

    @classmethod
    def show_message(cls, msg: str, *args, **kwargs):
        """Show an message to the user and log it

        Args:
            msg (str): User message to print on the console
        """
        click.secho(msg, fg="green")
        logging.info(msg, *args, **kwargs)

    def show_error(self, msg: str, *args, **kwargs):
        """Show an error to the user and log it

        Args:
            msg (str): User message to print on the console
        """
        if not self.verbose:
            click.secho(msg, fg="red", err=True)

        logging.warning(msg, *args, **kwargs)

    def show_info(self, msg: str, *args, **kwargs):
        """Show an info message to the user and log it

        Args:
            msg (str): User message to print on the console
        """
        if not self.verbose:
            click.secho(msg)

        logging.warning(msg, *args, **kwargs)

    def show_warning(self, msg: str, *args, **kwargs):
        """Show a warning to the user and log it

        Args:
            msg (str): User message to print on the console
        """
        if not self.verbose:
            click.secho(msg, fg="yellow", err=True)

        logging.warning(msg, *args, **kwargs)

    def set_env(self):
        """Set environment variables so information about the proxy can
        be access by plugins
        """
        os.environ["C8Y_HOST"] = str(self.host)
        os.environ["PORT"] = str(self.used_port)
        os.environ["DEVICE"] = self.device

        # Support WSL environments and expose variables to be explosed to WSL
        os.environ["WSLENV"] = "PORT/u:DEVICE/u:C8Y_HOST/u"


@dataclasses.dataclass
class RemoteAccessConnectionData:
    """Remote access connection data"""

    client: CumulocityClient
    managed_object_id: str
    remote_config_id: str


PASSTHROUGH = "PASSTHROUGH"
REMOTE_ACCESS_FRAGMENT = "c8y_RemoteAccessList"


class CliLogger:
    """CLI Logger"""

    # pylint: disable=too-few-public-methods

    @classmethod
    def log_path(cls) -> pathlib.Path:
        """Get the log path"""
        return (
            pathlib.Path(os.getenv("C8YLP_LOG_DIR", "~/.c8ylp/")).expanduser()
            / "localproxy.log"
        )


def configure_logger(path: pathlib.Path, verbose: bool = False) -> logging.Logger:
    """Configure logger

    Args:
        path (pathlib.Path): Path where the persistent logger should write to.
        verbose (bool, optional): Use verbose logging. Defaults to False.

    Returns:
        logging.Logger: Created logger
    """
    path.parent.mkdir(parents=True, exist_ok=True)

    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    log_file_formatter = logging.Formatter(
        "%(asctime)s %(threadName)s %(levelname)s %(name)s %(message)s"
    )

    # Set default log format
    if verbose:
        log_console_formatter = logging.Formatter(
            "[c8ylp]  %(levelname)-5s %(message)s"
        )
        console_loglevel = logging.INFO
        if len(logger.handlers) == 0:
            console_handler = logging.StreamHandler()
            console_handler.setFormatter(log_console_formatter)
            console_handler.setLevel(console_loglevel)
            logger.addHandler(console_handler)
        else:
            handler = logger.handlers[0]
            # ignore console log messages
            handler.setLevel(console_loglevel)
            handler.setFormatter(log_console_formatter)
    else:
        # Remove default console logging and only use file logging
        logger.handlers = []

    # Max 5 log files each 10 MB.
    rotate_handler = RotatingFileHandler(
        filename=str(path), maxBytes=10000000, backupCount=5
    )
    rotate_handler.setFormatter(log_file_formatter)
    rotate_handler.setLevel(logging.INFO)

    # Log to Rotating File
    logger.addHandler(rotate_handler)
    return logger


def signal_handler(_signal, _frame):
    """Signal handler"""
    sys.exit(ExitCodes.TERMINATE)


def register_signals():
    """Register signal handlers"""
    signal.signal(signal.SIGINT, signal_handler)


def create_client(ctx: click.Context, opts: ProxyContext) -> CumulocityClient:
    """Create Cumulocity client and prompt for missing credentials
    if necessary.

    Args:
        ctx (click.Context): Click context
        opts (ProxyContext): Proxy options

    Returns:
        CumulocityClient: Configured Cumulocity client
    """
    if not opts.disable_prompts and not opts.host:
        opts.host = click.prompt(
            text="Enter the Cumulocity Host/URL",
        )

    client = CumulocityClient(
        hostname=opts.host,
        tenant=opts.tenant,
        user=opts.user,
        password=opts.password,
        tfacode=opts.tfa_code,
        token=opts.token,
        ignore_ssl_validate=opts.ignore_ssl_validate,
    )

    if not client.url:
        opts.show_error(
            "No Cumulocity host was provided. The host can be set via"
            "environment variables, arguments or the env-file"
        )
        ctx.exit(ExitCodes.NO_SESSION)

    logging.info("Checking tenant id")
    client.validate_tenant_id()

    # Retry logging (except for stdio) so the user can be prompted for
    # their credentials/TFA code etc. without having to run c8ylp again.
    retries = 1 if opts.stdio else 3
    success = False
    while retries:
        try:
            if client.token:
                client.validate_credentials()
            else:
                client.login()

            if opts.env_file and opts.store_token:
                store_credentials(opts, client)

            success = True
            break
        except CumulocityMissingTFAToken:
            if not opts.disable_prompts:
                client.tfacode = click.prompt(
                    text="Enter your Cumulocity TFA-Token", hide_input=False
                )
        except Exception as ex:
            logging.info("unknown exception: %s", ex)

            if not opts.disable_prompts and not client.user:
                client.user = click.prompt(
                    text="Enter your Cumulocity Username",
                )

            if not opts.disable_prompts and not client.password:
                client.password = click.prompt(
                    text="Enter your Cumulocity Password [input hidden]",
                    hide_input=True,
                )

        retries -= 1

    if not success:
        logging.info("Could not create client")
        ctx.exit(ExitCodes.NO_SESSION)

    return client


def store_credentials(opts: ProxyContext, client: CumulocityClient):
    """Store credentials to the environment file. It creates
    the file if it does not already exist.

    The file will only be written to if it has changed.

    Args:
        opts (ProxyContext): Proxy options
        client (CumulocityClient): Cumulocity client containing valid
            credentials
    """
    changed = save_env(
        opts.env_file,
        {
            # Note: Don't save password!
            "C8Y_HOST": client.url,
            "C8Y_USER": client.user,
            "C8Y_TENANT": client.tenant,
            "C8Y_TOKEN": client.token,
        },
    )

    if changed:
        opts.show_message(f"Env file was updated: {opts.env_file}")
    else:
        opts.show_info(f"Env file is already up to date: {opts.env_file}")


def get_config_id(ctx: click.Context, mor: Dict[str, Any], config: str) -> str:
    """Get the remote access configuration id matching a specific type
    from a device managed object

    Args:
        mor (Dict[str, Any]): Device managed object
        config (str): Expected configuration name. If no matching name
            is found, but there are other PASSTHROUGH configuration available it will
            use the first config (and log a warning)

    Returns:
        str: Remote access configuration id
    """
    device_name = mor.get("name", "<<empty_name>>")
    if REMOTE_ACCESS_FRAGMENT not in mor:
        logging.error(
            'No Remote Access Configuration has been found for device "%s"', device_name
        )
        ctx.exit(ExitCodes.DEVICE_MISSING_REMOTE_ACCESS_FRAGMENT)

    valid_configs = [
        item
        for item in mor.get(REMOTE_ACCESS_FRAGMENT, [])
        if item.get("protocol") == PASSTHROUGH
    ]

    if not valid_configs:
        logging.error(
            'No config with protocol set to "%s" has been found for device "%s"',
            PASSTHROUGH,
            device_name,
        )
        ctx.exit(ExitCodes.DEVICE_NO_PASSTHROUGH_CONFIG)

    def extract_config_id(matching_config):
        logging.info(
            'Using Configuration with Name "%s" and Remote Port %s',
            matching_config.get("name"),
            matching_config.get("port"),
        )
        return matching_config.get("id")

    if not config:
        # use first config
        return extract_config_id(valid_configs[0])

    # find config matching name
    matches = [
        item
        for item in valid_configs
        if item.get("name", "").casefold() == config.casefold()
    ]

    if matches:
        return extract_config_id(matches[0])

    if valid_configs:
        # Fallback to using the first valid config (if not matches names were found)
        logging.warning(
            "No matching %s configs found (name=%s), so falling back to first config (name=%s)",
            PASSTHROUGH,
            config,
            valid_configs[0].get("name", ""),
        )
        return extract_config_id(valid_configs[0])

    # Technically it should never reach here, however leave in place in
    # case the above logic is changed
    logging.error(
        'Provided config name "%s" for "%s" was not found or none with protocol set to "%s"',
        config,
        device_name,
        PASSTHROUGH,
    )
    ctx.exit(ExitCodes.DEVICE_NO_MATCHING_PASSTHROUGH_CONFIG)


def run_proxy_in_background(
    ctx: click.Context,
    opts: ProxyContext,
    connection_data: RemoteAccessConnectionData,
    ready_signal: threading.Event = None,
):
    """Run the proxy in a background thread

    Args:
        ctx (click.Context): Click context
        opts (ProxyContext): Proxy options
        connection_data (RemoteAccessConnectionData): Remote access connection data
    """

    stop_signal = threading.Event()
    _local_ready_signal = threading.Event()

    # register signals as the proxy will be starting in a background thread
    # to enable the proxy to run as a subcommand
    register_signals()

    # Start the proxy in a background thread so the user can
    background = threading.Thread(
        target=start_proxy,
        args=(ctx, opts),
        kwargs={
            "connection_data": connection_data,
            "stop_signal": stop_signal,
            "ready_signal": _local_ready_signal,
        },
        daemon=True,
    )
    background.start()

    # Block until the local proxy is ready to accept connections
    if not _local_ready_signal.wait(opts.wait_port_timeout):
        opts.exit_server_not_ready()

    # Inject custom env variables for use within the script
    opts.set_env()

    # The subcommand is called after this
    timer = CommandTimer("Duration", on_exit=click.echo).start()

    # Shutdown the server once the plugin has been run
    @ctx.call_on_close
    def _shutdown_server_thread():
        stop_signal.set()
        background.join()
        timer.stop_with_message()

    # Only set ready signal once the whole env include env variables has
    # been setup
    if ready_signal:
        ready_signal.set()


def pre_start_checks(
    ctx: click.Context, opts: ProxyContext
) -> Optional[RemoteAccessConnectionData]:
    """Run prestart checks before starting the local proxy

    Args:
        ctx (click.Context): Click context
        opts (ProxyContext): Proxy options

    Returns:
        Optional[RemoteAccessConnectionData]: Remote access connection data
    """

    if opts.stdio and not sys.stdin.isatty():
        opts.disable_prompts = True

    try:
        client = create_client(ctx, opts)
        mor = client.get_managed_object(opts.device, opts.external_type)
        config_id = get_config_id(ctx, mor, opts.config)
        device_id = mor.get("id")

        is_authorized = client.validate_remote_access_role()
        if not is_authorized:
            opts.show_error(
                "The user is not authorized to use Cloud Remote Access. "
                f"Contact your Cumulocity Admin. user={opts.user}",
            )
            ctx.exit(ExitCodes.MISSING_ROLE_REMOTE_ACCESS_ADMIN)

    except Exception as ex:
        if isinstance(ex, click.exceptions.Exit):
            opts.show_error(f"Could not retrieve device information. reason={ex}")
            # re-raise existing exit
            raise

        error_context = ""
        extra_details = []
        if opts.host and opts.host not in str(ex):
            extra_details.append(f"host={opts.host or ''}")

        if opts.user and opts.user not in str(ex):
            extra_details.append(f"user={opts.user or ''}")

        if extra_details:
            error_context = ". settings: " + ", ".join(extra_details)

        opts.show_error(
            "Unexpected error when retrieving device information from Cumulocity. "
            f"error_details={ex}{error_context}"
        )
        ctx.exit(ExitCodes.NOT_AUTHORIZED)

    return RemoteAccessConnectionData(
        client=client, managed_object_id=device_id, remote_config_id=config_id
    )


def start_proxy(
    ctx: click.Context,
    opts: ProxyContext,
    connection_data: RemoteAccessConnectionData,
    stop_signal: threading.Event = None,
    ready_signal: threading.Event = None,
) -> NoReturn:
    """Start the local proxy

    Args:
        ctx (click.Context): Click context
        opts (ProxyContext): Proxy options
    """
    # pylint: disable=too-many-branches,too-many-statements
    is_main_thread = threading.current_thread() is threading.main_thread()
    if is_main_thread:
        register_signals()

    client_opts = {
        "host": opts.host,
        "config_id": connection_data.remote_config_id,
        "device_id": connection_data.managed_object_id,
        "session": connection_data.client.session,
        "token": opts.token,
        "ignore_ssl_validate": opts.ignore_ssl_validate,
        "ping_interval": opts.ping_interval,
    }

    socket_server = None
    background = None
    exit_code = ExitCodes.OK

    try:
        if opts.socket_path:
            socket_server = UnixStreamProxyServer(
                opts.socket_path,
                WebsocketClient(**client_opts),
                opts.tcp_size,
                opts.tcp_timeout,
            )
        elif opts.stdio:
            socket_server = StdioProxyServer(
                WebsocketClient(**client_opts),
                opts.tcp_size,
            )
        else:
            socket_server = TCPProxyServer(
                opts.port,
                WebsocketClient(**client_opts),
                opts.tcp_size,
                opts.tcp_timeout,
            )

        click.secho(BANNER1)
        logging.info("Starting socket server")

        background = threading.Thread(target=socket_server.serve_forever, daemon=True)
        background.start()

        # Block until the local proxy is ready to accept connections
        if not socket_server.wait_for_running(opts.wait_port_timeout):
            opts.exit_server_not_ready()

        # store the used port for reference to later
        if socket_server.server.socket:
            opts.used_port = socket_server.server.socket.getsockname()[1]

        # Plugins start in a background thread so don't display it
        # as the plugins should do their own thing
        if is_main_thread:
            opts.show_info(
                f"\nc8ylp is listening for device (ext_id) {opts.device} ({opts.host}) on localhost:{opts.used_port}",
            )
            ssh_username = opts.ssh_user or "<device_username>"
            msg = (
                f"\nFor example, if you are running a ssh proxy, you connect to {opts.device} by executing the "
                "following in a new tab/console:\n\n"
            )
            if opts.socket_path:
                msg += f"\tssh -o 'ProxyCommand=socat - UNIX-CLIENT:{opts.socket_path}' {ssh_username}@localhost"
            else:
                msg += f"\tssh -p {opts.used_port} {ssh_username}@localhost"
            opts.show_message(msg)

            opts.show_info("\nPress ctrl-c to shutdown the server")

        if ready_signal:
            ready_signal.set()

        # loop, waiting for server to stop
        while background.is_alive():
            if stop_signal and stop_signal.is_set():
                break
            time.sleep(1)
            logging.debug(
                "Waiting in background: alive=%s",
                background.is_alive(),
            )
    except Exception as ex:
        if isinstance(ex, click.exceptions.Exit):
            # propagate exit code
            exit_code = getattr(ex, "exit_code")
            raise

        if str(ex):
            opts.show_error(
                "The local proxy socket Server experienced an unexpected error. "
                f"port={opts.port}, error={ex}"
            )
            exit_code = ExitCodes.UNKNOWN
    finally:
        if socket_server:
            socket_server.shutdown()

        if background:
            background.join()

        if is_main_thread:
            if int(exit_code) == 0:
                opts.show_message(f"Exiting: {str(exit_code)} ({int(exit_code)})")
            else:
                opts.show_error(f"Exiting: {str(exit_code)} ({int(exit_code)})")

            ctx.exit(exit_code)
        else:
            opts.show_info("Exiting")



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""TCP server"""

import errno
import logging
import socket
import socketserver
import threading


class TCPHandler(socketserver.BaseRequestHandler):
    """
    The request handler class for our server.

    It is instantiated once per connection to the server
    """

    def handle(self):
        """Websocket connection handler"""
        request: socket.socket = self.request

        laddr = request.getsockname()
        raddr = request.getpeername()
        logging.info(
            "Handling request: fd=%s, laddr=%s, raddr=%s",
            request.fileno(),
            laddr,
            raddr,
        )

        def handle_shutdown():
            # Force shutdown of any socket reads or writes
            # Note: fileno is set to -1 if the socket has been closed
            if request.fileno() != -1:
                request.shutdown(socket.SHUT_RDWR)
                request.close()

        self.server.web_socket_client.shutdown_request = handle_shutdown

        # connect websocket
        if not self.server.web_socket_client.is_open():
            self.server.web_socket_client.connect()

        # update link to current tcp socket send
        request_error = False

        def safe_send(data):
            nonlocal request_error
            try:
                if not request_error:
                    request.send(data)
            except OSError as ex:
                request_error = True
                if ex.errno == errno.EBADF:
                    # TODO: Check what is sending messages after it has been closed
                    logging.info("Ignoring Bad file descriptor")
                else:
                    logging.error(ex)

        self.server.web_socket_client.proxy_send_message = safe_send

        while True:
            try:
                logging.debug("Reading from tcp port")
                data = request.recv(self.server.buffer_size or 1024)
                logging.debug("%s wrote: %s", self.client_address, data)

                if not data:
                    logging.debug("No data. Request will be closed")
                    self.server.web_socket_client.proxy_send_message = None
                    self.server.web_socket_client.stop()
                    break

                logging.debug("Writing data to ws: %s", data)
                self.server.web_socket_client.send_binary(data)
                logging.debug("Wrote data to ws: %s", data)
            except ConnectionResetError as ex:
                logging.info("Connection was reset. %s", ex)
                break


class CustomTCPServer(socketserver.TCPServer):
    """Custom TCP Server used to listen for local connections and proxy them to
    a websocket
    """

    def __init__(
        self,
        web_socket_client,
        server_address,
        RequestHandlerClass,
        bind_and_activate=True,
        buffer_size: int = 1024,
        tcp_timeout: float = 30,
        port: int = None,
    ) -> None:
        self.allow_reuse_address = True
        self.port = port
        self.timeout = tcp_timeout
        self.buffer_size = buffer_size
        self.web_socket_client = web_socket_client
        super().__init__(
            server_address, RequestHandlerClass, bind_and_activate=bind_and_activate
        )


class TCPProxyServer:
    """TCP Server"""

    def __init__(
        self,
        port,
        web_socket_client,
        tcp_buffer_size,
        tcp_timeout,
    ):
        self.web_socket_client = web_socket_client
        self._running = threading.Event()
        self.logger = logging.getLogger(__name__)

        # Expose func to web socket client
        self.web_socket_client.proxy = self

        self.server = CustomTCPServer(
            self.web_socket_client,
            ("localhost", port),
            TCPHandler,
            port=port,
            buffer_size=tcp_buffer_size,
            tcp_timeout=tcp_timeout,
        )

    def serve_forever(self):
        """Server tcp server forever. Only returns once .shutdown has been called"""
        self._running.set()
        try:
            self.server.serve_forever()
        except Exception as ex:
            logging.error("TCP Server raised an exception. %s", ex, exc_info=True)
            raise ex
        self._running.clear()

    def shutdown(self):
        """Shutdown tcp server"""
        if self._running.is_set():
            logging.info("Shutting down TCP server")

            if callable(self.server.web_socket_client.shutdown_request):
                self.server.web_socket_client.shutdown_request()
            self.server.shutdown()

    def wait_for_running(self, timeout: float = 30.0) -> bool:
        """Wait for the server to startup

        Args:
            timeout (float, optional): Timeout in seconds. Defaults to 30.0.

        Returns:
            bool: True if the server is running
        """
        return self._running.wait(timeout)



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Cumulocity client"""
#  Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
#  SPDX-License-Identifier: Apache-2.0
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import json
import logging
import urllib.parse
from typing import Any, Dict

import requests
from requests.auth import HTTPBasicAuth

from c8ylp.rest_client.sessions import BaseUrlSession


class CumulocityError(Exception):
    """Cumulocity error"""


class CumulocityMissingTFAToken(Exception):
    """Cumulocity missing TFA token"""


class CumulocityPermissionDeviceError(Exception):
    """Cumulocity device permission error"""

    def __init__(self, user: str, device: str, host: str) -> None:
        message = f"User {user} is not authorized to read Device Data of Device {device} on {host}"
        super().__init__(message)


class CumulocityDeviceNotFound(Exception):
    """Cumulocity device not found error"""

    def __init__(self, device: str, host: str) -> None:
        message = f"A device with the external id '{device}' was not found in Cumulocity host: {host}"
        super().__init__(message)


class BearerAuth(requests.auth.AuthBase):
    """Bearer/token based authorization"""

    # pylint: disable=too-few-public-methods
    def __init__(self, token: str):
        self.token = token

    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.token
        return r


class CumulocityClient:
    """Cumulocity REST Client"""

    def __init__(
        self,
        hostname: str,
        tenant: str = "",
        user: str = "",
        password: str = "",
        tfacode: str = "",
        token: str = "",
        ignore_ssl_validate: bool = False,
    ):
        hostname = hostname or ""

        if hostname.startswith("http"):
            self.url = hostname
        else:
            self.url = f"https://{hostname}"

        self.session = BaseUrlSession(prefix_url=self.url)

        if ignore_ssl_validate:
            self.session.verify = False

        self.tenant = tenant
        self.user = user
        self.password = password
        self.tfacode = tfacode
        self.token = token
        self._supports_oauth = False

        if token:
            self.set_bearer_auth(token)
        else:
            self.set_basic_auth(user, password)

        self.logger = logging.getLogger(__name__)

    def validate_tenant_id(self) -> str:
        """Validate the tenant id configured in the client
        If it is does not match the Url, then it will be updated to match.

        Returns:
            str: Tenant ID
        """
        tenant_id = ""
        response = self.session.get("/tenant/loginOptions")
        if response.status_code == 200:
            login_options_body = json.loads(response.content.decode("utf-8"))
            login_options = login_options_body.get("loginOptions", {})
            for option in login_options:
                # Check if oAuth is supported
                if option.get("type") in ("OAUTH2_INTERNAL", "OAUTH2"):
                    self._supports_oauth = True

                if "initRequest" in option:
                    _, _, query_str = option.get("initRequest", "").partition("?")
                    query_params = urllib.parse.parse_qs(query_str)

                    tenant_id = (
                        query_params["tenant_id"][0]
                        if "tenant_id" in query_params
                        else ""
                    )
                    if not tenant_id:
                        continue

                    if self.tenant != tenant_id:
                        if self.tenant:
                            # Skip warning if the tenant id is not set
                            self.logger.warning(
                                "Wrong Tenant ID %s, Correct Tenant ID: %s",
                                self.tenant,
                                tenant_id,
                            )
                        self.tenant = tenant_id
                    else:
                        tenant_id = ""
                    break

        else:
            self.logger.error("Could not validate Tenant ID")
        return tenant_id

    def validate_remote_access_role(self) -> bool:
        """Check if the current user has permission to access the remote access feature
        i.e. the ROLE_REMOTE_ACCESS_ADMIN role.

        Returns:
            bool: True if the user had the correct role assigned to them
        """
        is_valid = False
        response = self.session.get("/user/currentUser")
        if response.status_code == 200:
            user = json.loads(response.content.decode("utf-8"))
            effective_roles = user["effectiveRoles"]
            for role in effective_roles:
                if role.get("id") == "ROLE_REMOTE_ACCESS_ADMIN":
                    self.logger.debug(
                        "Remote Access Role assigned to User %s", self.user
                    )
                    is_valid = True
                    break
        else:
            self.logger.error("Could not get the user's data")
            is_valid = False
        return is_valid

    def set_bearer_auth(self, token: str) -> None:
        """Set authorization header to use a bearer token

        Args:
            token (str): [description]
        """
        self.session.auth = BearerAuth(token)

    def set_basic_auth(self, username: str, password: str) -> None:
        """Set the authorization header to use basic authentication

        Args:
            username (str): Cumulocity username
            password (str): Cumulocity password
        """
        self.session.auth = HTTPBasicAuth(username or "", password or "")

    def validate_credentials(self) -> bool:
        """Validate client's credentials by sending a request

        If the request fails, then the token and password will be configured
        in the client will be set to None.

        Raises:
            PermissionError: Permission error if the credentials are not valid

        Returns:
            bool: True if the credentials were validated
        """
        response = self.session.get("/user/currentUser")
        if response.status_code == 200:
            return True

        message = f"Error validating Token {response.status_code}. Please provide Tenant, User and Password"

        if response.headers.get("content-type") == "application/json":
            response_json = response.json()
            if "message" in response_json:
                message = response_json()

        # clear token/password (as they maybe invalid)
        self.token = ""
        self.password = ""

        error = PermissionError(message)
        self.logger.error(error)
        raise error

    def login(self):
        """Login or validate Cumulocity Credentials.
        For OAuth type logins, a token will be generated (The perferred method).

        If the tenant only has Basic Auth enabled, then the given credentials will
        be validated.
        """
        if self._supports_oauth:
            self.login_oauth()
        else:
            # Basic AUTH does not required login, but validate
            # the credentials
            self.validate_credentials()

    def login_oauth(self):
        """Login via OAuth2 and set the give token if the login is successful

        Raises:
            PermissionError: Error if the provided credentials and/or tfa code are incorrect
            Exception: Unexpected error
        """
        oauth_url = f"/tenant/oauth?tenant_id={self.tenant}"
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
        }
        body = {
            "grant_type": "PASSWORD",
            "username": self.user,
            "password": self.password,
            "tfa_code": self.tfacode,
        }
        self.logger.debug(
            "Sending requests to %s with body %s",
            oauth_url,
            str(body).replace(str(self.password), "******"),
        )
        self.session.auth = None
        response = self.session.post(oauth_url, headers=headers, data=body)
        if response.status_code == 200:
            self.logger.debug(
                "Authentication successful. Tokens have been updated %s",
                self.session.cookies.get_dict(),
            )
            self.token = self.session.cookies.get_dict()["authorization"]
            self.set_bearer_auth(self.token)
            self.session.headers["X-XSRF-TOKEN"] = self.session.cookies.get_dict()[
                "XSRF-TOKEN"
            ]
        elif response.status_code == 401:
            # Check if TFA is required
            if "TFA" in response.text and not self.tfacode:
                error = CumulocityMissingTFAToken()
            else:
                error = PermissionError(
                    response.json().get(
                        "message", "User is not authorized to access Tenant"
                    ),
                    # f"User {self.user} is not authorized to access Tenant {self.tenant} or TFA-Code is invalid.",
                )
                self.logger.error(error)
            raise error
        else:
            error = Exception(
                f"Server Error received for User {self.user} and Tenant {self.tenant}. "
                f"Status Code: {response.status_code}"
            )
            self.logger.error(error)
            raise error

    def get_external_id(
        self, serial_number: str, identity_type: str = "c8y_Serial"
    ) -> Dict[str, Any]:
        """Get the external id provided the serial number and external id type

        Args:
            serial_number (str): Device external identity
            identity_type (str, optional): External identity type. Defaults to 'c8y_Serial'.

        Raises:
            CumulocityPermissionDeviceError: Error when the user does not have the correct permissions
                to access the API or device
            Exception: Unexpected error

        Returns:
            Dict[str, Any]: External Id definition containing the managed object id
        """
        response = self.session.get(
            f"/identity/externalIds/{identity_type}/{serial_number}"
        )

        if response.status_code == 200:
            return json.loads(response.content.decode("utf-8"))

        error = Exception(
            f"Error retrieving device. Status Code {response.status_code}"
        )
        if response.status_code == 401:
            error = CumulocityPermissionDeviceError(self.user, serial_number, self.url)
        elif response.status_code == 404:
            error = CumulocityDeviceNotFound(serial_number, self.url)

        self.logger.error(error)
        raise error

    def get_managed_object(
        self, serial_number: str, identity_type: str = "c8y_Serial"
    ) -> Dict[str, Any]:
        """Get a managed object by looking it up via its external identity

        Args:
            serial_number (str): Device external identity
            identity_type (str, optional): External identity type. Defaults to 'c8y_Serial'.

        Raises:
            CumulocityPermissionDeviceError: Error when the user does not have the correct permissions
                to access the API or device
            Exception: Unexpected error

        Returns:
            Dict[str, Any]: Device managed object
        """
        ext_id = self.get_external_id(serial_number, identity_type)
        mor_id = ext_id.get("managedObject", {}).get("id")

        if not mor_id:
            raise CumulocityError("Managed object id is empty")

        response = self.session.get(f"/inventory/managedObjects/{mor_id}")
        if response.status_code == 200:
            return json.loads(response.content.decode("utf-8"))

        error = Exception(
            f"Error retrieving device. Status Code {response.status_code}, "
            f"device (external id)={serial_number}, host={self.url}, user={self.user}"
        )
        if response.status_code == 401:
            error = CumulocityPermissionDeviceError(self.user, serial_number, self.url)
        elif response.status_code == 404:
            error = CumulocityDeviceNotFound(serial_number, self.url)

        self.logger.error(error)
        raise error



#
# Copyright (c) 2022 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Stdin/Stdout forwarding server"""
import io
import logging
import sys
import threading

from c8ylp.tcp_socket.tcp_server import TCPProxyServer


class StdioHandler:
    """
    The request handler class for our server.

    It is instantiated once and used to forward stdin/stdout to the websocket
    """

    def __init__(self, web_socket_client, buffer_size):
        self.web_socket_client = web_socket_client
        self.buffer_size = buffer_size
        self.shutdown_event = threading.Event()
        self.socket = None
        self._reader = None
        self._writer = None

    def shutdown(self):
        """Shutdown tcp server"""
        self.shutdown_event.set()
        self._reader.close()
        self._writer.close()

    def serve_forever(self):
        """stdio "connection" handler"""
        with io.open(
            sys.stdin.fileno(), mode="rb", closefd=False, buffering=0
        ) as self._reader, io.open(
            sys.stdout.fileno(), mode="wb", closefd=False, buffering=0
        ) as self._writer:
            self.web_socket_client.shutdown_request = self.shutdown

            # connect websocket
            if not self.web_socket_client.is_open():
                self.web_socket_client.connect()

            # update link to current stdin/stdout send
            request_error = False

            def safe_send(data):
                nonlocal request_error
                try:
                    if not request_error:
                        self._writer.write(data)
                except OSError as ex:
                    request_error = True
                    logging.error(ex)

            self.web_socket_client.proxy_send_message = safe_send

            while not self.shutdown_event.is_set():
                try:
                    logging.debug("Reading from stdin")
                    data = self._reader.read(self.buffer_size or 1024)
                    logging.debug("stdin wrote: %s", data)
                    if not data:
                        logging.debug("No data. Request will be closed")
                        self.web_socket_client.proxy_send_message = None
                        self.web_socket_client.stop()
                        break

                    logging.debug("Writing data to ws: %s", data)
                    self.web_socket_client.send_binary(data)
                    logging.debug("Wrote data to ws: %s", data)
                except (ConnectionResetError, OSError) as ex:
                    logging.info("Connection was reset. %s", ex)
                    break


class StdioProxyServer(TCPProxyServer):
    """
    Stdin/Stdout forwarding server
    """

    # pylint: disable=super-init-not-called
    def __init__(
        self,
        web_socket_client,
        buffer_size,
    ):
        self.web_socket_client = web_socket_client
        self._running = threading.Event()
        self.logger = logging.getLogger(__name__)

        # Expose func to web socket client
        self.web_socket_client.proxy = self

        self.server = StdioHandler(self.web_socket_client, buffer_size)



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""SSH command tests"""
import pathlib
from unittest.mock import Mock, patch

import pytest
import responses
from click.testing import CliRunner
from c8ylp.cli.core import PASSTHROUGH, REMOTE_ACCESS_FRAGMENT, ExitCodes
from c8ylp.main import cli
from tests.env import Environment
from tests.fixtures import FixtureCumulocityAPI


@patch("subprocess.call", return_value=0)
def test_single_ssh_command_then_exit(
    mock_start_ssh: Mock, c8yserver: FixtureCumulocityAPI, env: Environment
):
    """Execute command via ssh then exit"""

    @responses.activate
    def run():
        serial = "ext-device-01"
        c8yserver.simulate_pre_authenticated(serial)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["connect", "ssh", serial, "--ssh-user", "example", "ls -la"],
            env=env.create_authenticated(),
        )

        mock_start_ssh.assert_called_once()
        assert result.exit_code == 0
        assert result.stdout

    run()


@patch("shutil.which", return_value=None)
def test_single_ssh_command_with_missing_ssh_binary(
    _mock_which: Mock, c8yserver: FixtureCumulocityAPI, env: Environment
):
    """Execute command via ssh then exit"""

    @responses.activate
    def run():
        serial = "ext-device-01"
        c8yserver.simulate_pre_authenticated(serial)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["connect", "ssh", serial, "--ssh-user", "example", "ls -la"],
            env=env.create_authenticated(),
        )

        assert result.exit_code == ExitCodes.SSH_NOT_FOUND

    run()


# @patch("c8ylp.plugins.ssh.subprocess.call", return_code=0)
@patch("subprocess.call", return_value=0)
def test_launching_ssh_with_fixed_port(
    mock_call: Mock, c8yserver: FixtureCumulocityAPI, env: Environment
):
    """Execute command via ssh client using a specific port"""

    @responses.activate
    def run():
        serial = "ext-device-01"
        port = "1234"
        username = "admin"
        c8yserver.simulate_pre_authenticated(serial)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "connect",
                "ssh",
                serial,
                "--port",
                port,
                "--ssh-user",
                username,
                "ls -la",
            ],
            env=env.create_authenticated(),
        )

        assert result.exit_code == ExitCodes.OK
        mock_call.assert_called_once()
        ssh_cmd = mock_call.call_args[0][0]

        assert ssh_cmd == [
            "ssh",
            "-o",
            "ServerAliveInterval=120",
            "-o",
            "StrictHostKeyChecking=no",
            "-o",
            "UserKnownHostsFile=/dev/null",
            "-p",
            port,
            f"{username}@localhost",
            "ls -la",
        ]

        assert mock_call.call_args[1]["env"]

    run()


@patch("subprocess.call", return_value=99)
def test_return_exit_code(
    mock_start_ssh: Mock, c8yserver: FixtureCumulocityAPI, env: Environment
):
    """Execute ssh command and resulting exit code is returned via c8ylp"""

    @responses.activate
    def run():
        serial = "ext-device-01"
        c8yserver.simulate_pre_authenticated(serial)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "connect",
                "ssh",
                serial,
                "--verbose",
                "--ssh-user",
                "example",
                "ls -l /etc; exit 99",
            ],
            env=env.create_authenticated(),
        )
        assert result.exit_code == 99
        assert result.stdout

        mock_start_ssh.assert_called_once()
        ssh_args = mock_start_ssh.call_args[0][0]
        assert "ls -l /etc; exit 99" in ssh_args
        assert "example@localhost" in ssh_args

    run()


@patch("subprocess.call", return_value=0)
def test_prompt_for_ssh_user(
    mock_start_ssh, c8yserver: FixtureCumulocityAPI, env: Environment
):
    """Prompts for ssh user if it is not set via cli or env"""

    @responses.activate
    def run():
        serial = "ext-device-01"
        c8yserver.simulate_pre_authenticated(serial_number=serial)

        runner = CliRunner()
        test_env = {
            **env.create_authenticated(),
            "C8YLP_SSH_USER": "",
        }
        result = runner.invoke(
            cli, ["connect", "ssh", serial, "ls -la"], env=test_env, input="admin"
        )

        mock_start_ssh.assert_called_once()
        assert result.exit_code == 0
        assert result.stdout
        ssh_args = mock_start_ssh.call_args[0][0]
        assert "admin@localhost" in ssh_args

    run()


@pytest.mark.parametrize(
    "inputs",
    (
        {
            "stdin": [
                "12345\n",
            ],
            "env": {
                "C8Y_HOST": "https://example.c8y.io",
            },
        },
        {
            "stdin": [],
            "env": {
                "C8Y_HOST": "https://example.c8y.io",
                "C8Y_TFA_CODE": "000000",
            },
        },
        {
            "stdin": [],
            "env": {
                # Url has a trialing slash
                "C8Y_HOST": "https://example.c8y.io/",
                "C8Y_TFA_CODE": "000000",
            },
        },
        {
            "stdin": [],
            "env": {
                # Url has whitespace and trialing slashes
                "C8Y_HOST": "  https://example.c8y.io////  ",
                "C8Y_TFA_CODE": "000000",
            },
        },
    ),
)
def test_prompt_for_tfa(
    inputs, c8yserver: FixtureCumulocityAPI, env: Environment, tmpdir
):
    """User is prompted for TFA code if the token is not provided"""

    @responses.activate
    def run():
        serial = "ext-device-01"
        c8yserver.simulate_loginoptions(tenant="t12345")
        c8yserver.simulate_login_oauth(status_codes=[401, 200])
        c8yserver.simulate_current_user()
        c8yserver.simulate_external_identity(serial_number=serial)
        c8yserver.simulate_managed_object()

        env_file = tmpdir.join(".env")
        pathlib.Path(env_file).touch()
        stdin = "".join(inputs["stdin"])

        with patch("subprocess.call", return_value=0):
            runner = CliRunner()
            result = runner.invoke(
                cli,
                [
                    "connect",
                    "ssh",
                    serial,
                    "--env-file",
                    env_file.strpath,
                    "--ssh-user",
                    "admin",
                ],
                env={
                    "C8Y_TENANT": "t12345",
                    "C8Y_USER": "example-user",
                    "C8Y_PASSWORD": "d4mmy-p4s$wurd",
                    "C8Y_TOKEN": "",
                    **inputs["env"],
                },
                input=stdin,
            )

        assert result.exit_code == 0

        settings = env.read_file(env_file)

        # Password should not be stored, only the token
        assert settings == {
            "C8Y_HOST": "https://example.c8y.io",
            "C8Y_USER": "example-user",
            "C8Y_TENANT": "t12345",
            "C8Y_TOKEN": "dummy-token-xyz",
        }

    run()


def test_missing_role(c8yserver: FixtureCumulocityAPI, env: Environment):
    """Test: Command should fail if the user does not have the correct ROLE"""

    @responses.activate
    def run():
        serial = "ext-device-01"
        c8yserver.simulate_pre_authenticated(serial_number=serial, roles=[])

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "connect",
                "ssh",
                serial,
                "--ssh-user",
                "admin",
            ],
            env=env.create_authenticated(),
        )

        assert result.exit_code == ExitCodes.MISSING_ROLE_REMOTE_ACCESS_ADMIN

    run()


@pytest.mark.parametrize(
    "case",
    (
        {
            "fragments": {},
            "exit_code": ExitCodes.DEVICE_MISSING_REMOTE_ACCESS_FRAGMENT,
        },
        {
            "description": "Missing PASSTHROUGH config",
            "fragments": {
                REMOTE_ACCESS_FRAGMENT: [
                    {"id": 1, "name": "example-ssh", "protocol": "ssh"}
                ]
            },
            "exit_code": ExitCodes.DEVICE_NO_PASSTHROUGH_CONFIG,
        },
        {
            "description": "Not an exact PASSTHROUGH name match",
            "fragments": {
                REMOTE_ACCESS_FRAGMENT: [
                    {"id": 1, "name": "example-ssh", "protocol": "ssh"},
                    {"id": 2, "name": "custom-passthrough", "protocol": PASSTHROUGH},
                ]
            },
            "exit_code": ExitCodes.MISSING_ROLE_REMOTE_ACCESS_ADMIN,
        },
        {
            "description": "Missing matching PASSTHROUGH name",
            "fragments": {
                REMOTE_ACCESS_FRAGMENT: [
                    {"id": 1, "name": "example-ssh", "protocol": "ssh"},
                ]
            },
            "exit_code": ExitCodes.DEVICE_NO_PASSTHROUGH_CONFIG,
        },
        {
            "description": "Custom PASSTHROUGH name matching but still missing role",
            "fragments": {
                REMOTE_ACCESS_FRAGMENT: [
                    {"id": 1, "name": "example-ssh", "protocol": "ssh"},
                    {"id": 2, "name": "custom-passthrough", "protocol": PASSTHROUGH},
                ]
            },
            "options": ["--config", "custom-passthrough"],
            "exit_code": ExitCodes.MISSING_ROLE_REMOTE_ACCESS_ADMIN,
        },
        {
            "description": "Custom PASSTHROUGH name matching but still missing role",
            "fragments": {
                REMOTE_ACCESS_FRAGMENT: [
                    {"id": 1, "name": "example-ssh", "protocol": "ssh"},
                    {"id": 2, "name": "custom-passthrough", "protocol": PASSTHROUGH},
                ]
            },
            "options": ["--config", ""],
            "exit_code": ExitCodes.MISSING_ROLE_REMOTE_ACCESS_ADMIN,
        },
    ),
)
def test_device_managed_object_permission_denied(
    case, c8yserver: FixtureCumulocityAPI, env: Environment
):
    """Test: Command should fail if the user does not have permission to read
    the device managed object
    """

    @responses.activate
    def run():
        serial = "ext-device-01"
        c8yserver.simulate_pre_authenticated(
            serial_number=serial,
            # skip roles
            roles=[],
            device_managed_object=case["fragments"],
        )

        runner = CliRunner()
        args = [
            "connect",
            "ssh",
            serial,
            *case.get("options", []),
            "--ssh-user",
            "admin",
        ]

        result = runner.invoke(
            cli,
            args,
            env=env.create_authenticated(),
        )

        assert result.exit_code == case["exit_code"], case["description"]

    run()


@patch("subprocess.call", return_value=0)
def test_deprecated_reconnects_option(
    mock_start_ssh: Mock, c8yserver: FixtureCumulocityAPI, env: Environment
):
    """Check deprecated warning when using reconnects option"""

    @responses.activate
    def run():
        serial = "ext-device-01"
        c8yserver.simulate_pre_authenticated(serial)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "connect",
                "ssh",
                serial,
                "--ssh-user",
                "example",
                "ls -la",
                "--reconnects",
                "2",
            ],
            env=env.create_authenticated(),
        )

        mock_start_ssh.assert_called_once()
        assert result.exit_code == 0
        assert result.stdout
        assert "Warning: 'reconnects' option is deprecated" in result.stdout

    run()


@patch("subprocess.call", return_value=0)
def test_deprecated_reconnects_option_when_not_used(
    mock_start_ssh: Mock, c8yserver: FixtureCumulocityAPI, env: Environment
):
    """Check deprecated warning when using reconnects option when it is not used"""

    @responses.activate
    def run():
        serial = "ext-device-01"
        c8yserver.simulate_pre_authenticated(serial)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["connect", "ssh", serial, "--ssh-user", "example", "ls -la ."],
            env=env.create_authenticated(),
        )

        mock_start_ssh.assert_called_once()
        assert result.exit_code == 0
        assert result.stdout
        assert "Warning: 'reconnects' option is deprecated" not in result.stdout

    run()


@patch("subprocess.call", return_value=0)
def test_reading_ssh_user_from_env(
    mock_start_ssh: Mock, c8yserver: FixtureCumulocityAPI, env: Environment
):
    """Read ssh user from an environment variable"""

    @responses.activate
    def run():
        serial = "ext-device-01"
        c8yserver.simulate_pre_authenticated(serial)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["connect", "ssh", serial, "ls -la ."],
            env={
                **env.create_authenticated(),
                "C8YLP_SSH_USER": "example",
            },
        )

        mock_start_ssh.assert_called_once()
        assert result.exit_code == 0
        assert result.stdout

    run()



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""SCP (Secure Copy) """

import filecmp
import logging
import subprocess
import sys
import time
import threading
from typing import List

import pytest

from .fixtures import Device, FileFactory


def proxy_cli(*args) -> subprocess.CompletedProcess:
    """Execute the proxy cli command with given arguments"""
    return subprocess.run(
        [
            sys.executable,
            "-m",
            "c8ylp",
            *args,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )


@pytest.mark.parametrize(
    "case",
    (
        {
            "size_mb": 1,
        },
    ),
    ids=lambda x: str(x),
)
def test_scp_file_transfer_roundtrip(
    case, c8ydevice: Device, file_factory: FileFactory
):
    """Test copy a file via scp to a remote device"""
    size_mb = case["size_mb"]
    name_prefix = f"dummyfile_{size_mb}"
    local_file = file_factory.create_file(name_prefix, size_mb)

    output = proxy_cli(
        "plugin",
        "command",
        c8ydevice.device,
        "--env-file",
        ".env",
        "--",
        "scp",
        "-o",
        "StrictHostKeyChecking=no",
        "-o",
        "UserKnownHostsFile=/dev/null",
        "-P",
        "$PORT",
        str(local_file),
        f"{c8ydevice.ssh_user}@localhost:",
    )

    assert output.returncode == 0

    # Get file from device, and compare with the sent file
    roundtrip_file = file_factory.dir / f"{name_prefix}_roundtrip"
    output = proxy_cli(
        "plugin",
        "command",
        c8ydevice.device,
        "--",
        "scp",
        "-o",
        "StrictHostKeyChecking=no",
        "-o",
        "UserKnownHostsFile=/dev/null",
        "-P",
        "$PORT",
        f"{c8ydevice.ssh_user}@localhost:/home/{c8ydevice.ssh_user}/{local_file.name}",
        str(roundtrip_file),
    )

    assert output.returncode == 0

    assert filecmp.cmp(local_file, roundtrip_file)


@pytest.mark.parametrize(
    "case",
    (
        dict(clients=1, delay=5, size_mb=10),
        dict(clients=2, delay=10, size_mb=10),
        dict(clients=2, delay=10, size_mb=100),
    ),
    ids=lambda x: str(x),
)
def test_concurrent_scp_commands(case, c8ydevice: Device, file_factory: FileFactory):
    clients = case.get("clients")
    delay = case.get("delay", 0.1)
    size_mb = case.get("size_mb", 10)
    name_prefix = f"concurrent_scp_{size_mb}"

    current_file = file_factory.create_file(name_prefix, size_mb)

    threads: List[threading.Thread] = []
    results: List[subprocess.CompletedProcess] = [None] * clients
    durations: List[float] = [None] * clients

    for i in range(0, clients):
        remote_file = f"{name_prefix}_client{i}.inttest"
        cmd_args = (
            "plugin",
            "command",
            c8ydevice.device,
            "--",
            "scp",
            "-o",
            "StrictHostKeyChecking=no",
            "-o",
            "UserKnownHostsFile=/dev/null",
            "-P",
            "$PORT",
            str(current_file),
            f"{c8ydevice.ssh_user}@localhost:{remote_file}",
        )

        def run_proxy(index, *args):
            start = time.monotonic()
            results[index] = proxy_cli(*args)
            durations[index] = time.monotonic() - start
            logging.info(
                f"client={index+1}, exit_code={results[index].returncode} duration={int(durations[index])}"
            )

        thread = threading.Thread(target=run_proxy, args=[i, *cmd_args], daemon=True)
        thread.start()
        threads.append(thread)

        if i < clients - 1:
            time.sleep(delay)

    # wait for threads to finish
    for thread in threads:
        thread.join()

    assert [proc.returncode for proc in results] == [0] * clients



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""login command"""

import click

from .. import options
from .core import ProxyContext, create_client


@click.command()
@options.HOSTNAME_PROMPT
@options.C8Y_TENANT
@options.C8Y_USER
@options.C8Y_TOKEN
@options.C8Y_PASSWORD
@options.C8Y_TFA_CODE
@options.LOGGING_VERBOSE
@options.ENV_FILE_OPTIONAL_EXISTS
@options.STORE_TOKEN
@options.DISABLE_PROMPT
@click.pass_context
def login(
    ctx: click.Context,
    *_args,
    **kwargs,
):
    """Login and save token to an environment file

    You will be prompted for all of the relevant information,
    i.e. host, username, password and TFA code (if required)

    Example 1: Create/update an env-file by trying to login into Cumulocity

    \b
        c8ylp login --env-file mytenant.env

    """
    opts = ProxyContext(ctx, kwargs)

    try:
        create_client(ctx, opts)
    except Exception:
        ctx.fail("Could not login")
